package it.sephiroth.scripting.utils
{
	public class RegexLexer implements ILexer
	{
		private static var eating: int = 0;
		
		private var _charno: int;
		private var _lineno: int;
		private var _columnno: int;
		private var _input: String;
		private var _current: Token;
		private var _source_name: String;
		
		private var _match_rx: RegExp;
		private var _rules: Array;
		
		public function RegexLexer( source_name: String, input: String )
		{
			_lineno = 1;
			_charno = 0;
			_columnno = 1;
			_rules = new Array();
			
			_input = input;
			_source_name = source_name;
		}
		
		public function get input(): String
		{
			return _input;
		}
		
		public function get lineno(): int
		{
			return _lineno;
		}
		
		public function get charno(): int
		{
			return _charno;
		}
		
		public function get columnno(): int
		{
			return _columnno;
		}
		
		public function get current(): Token
		{
			return _current;
		}
		
		public function get sourceName(): String
		{
			return _source_name;
		}
		
		public function addRule( name: String, pattern: String, type: int, callback: Function = null ): void
		{
			var index: int;
			var subpatterns: int;
			var rule: LexerMatchRule;
			var last_rule: LexerMatchRule;
			
			if( _rules.length > 0 )
			{
				last_rule = _rules[ ( _rules.length - 1) ];
				index = ( last_rule.index + last_rule.subpatterns + 1 );
			}else
			{
				index = 1;
			}
			
			subpatterns = countSubpatterns( pattern );
			
			rule = new LexerMatchRule();
			
			rule.type = type;
			rule.name = name;
			rule.index = index;
			rule.pattern = pattern;
			rule.subpatterns = subpatterns;
			rule.callback = callback;
			
			_rules.push( rule );
		}
		
		public function reset( input: String ): void
		{
			_lineno = 1;
			_charno = 0;
			_columnno = 1;
			_input = input;
			_current = null;
			_match_rx = null;
		}
		
		public function next(): Token
		{
			var token: Token;
			var match: Object;
			
			if( !_match_rx )
			{
				buildRegExp();
			}
			
			if( _match_rx.lastIndex >= _input.length )
			{
				token = new Token( Token.EOF, "$", _source_name, _charno, _lineno, _columnno );
			}else
			{
				match = _match_rx.exec( _input );

				if( match )
				{
					var i: int;
					var value: String;
					var subpatterns: Array;
					var rule: LexerMatchRule;
					var context: RegexLexerContext;
					
					if( match[ '_' ] )
					{
						throw new Error( "Unexpected character: " + _input.charAt( match.index ) + " at " + _source_name + ":" + _lineno + ":" + _columnno );
					}

					for each( rule in _rules )
					{
						if( match[ rule.name ] )
						{
							if( match[ rule.index ] == null )
							{
								throw new Error( "Internal error: inconsistent match!" );
							}

							value = match[ rule.name ];
							subpatterns = new Array();

							for( i = 1; i <= rule.subpatterns; ++i )
							{
								subpatterns.push( match[ ( rule.index + i ) ] );
							}
							
							token = new Token( rule.type, value, _source_name, _charno, _lineno, _columnno );
							
							if( rule.callback )
							{
								context = new RegexLexerContext();
								
								context.name = rule.name;
								context.lineno = _lineno;
								context.charno = _charno;
								context.columnno = _columnno;
								context.match = value;
								context.token = token;
								context.subpatterns = subpatterns;
								
								rule.callback( context );

								_charno += value.length;
								_columnno += value.length;
								
								if( _lineno != context.lineno )
								{
									_columnno = 1;
									_lineno = context.lineno;
								}
								
								token = context.token;
								
								if( context.skip || eating > 0 )
								{
									token = next();
								}
							}else
							{
								_charno += value.length;
								_columnno += value.length;
							}
						}
					}

					if( !token )
					{
						throw new Error( "Unexpected character: " + _input.charAt( match.index ) + " at " + _source_name + ":" + _lineno + ":" + _columnno );
					}
				}else
				{
					throw new Error( "Unexpected character: " + _input.charAt( _charno ) + " at " + _source_name + ":" + _lineno + ":" + _columnno );
				}
			}
			
			return token;
		}
		
		private function buildRegExp(): void
		{
			var pattern: String;
			var rule: LexerMatchRule;
			
			pattern = "";
			
			for each( rule in _rules )
			{
				pattern += "(?P<" + rule.name + ">" + rule.pattern + ")|";
			}
			
			pattern += "(?P<_>.)";
			
			_match_rx = new RegExp( pattern, "g" );
		}
		
		private function countSubpatterns( pattern: String ): int
		{
			pattern = pattern.replace( "\\(", "#" );
			return ( pattern.split( "(" ).length - 1 );
		}
		
		public static function SKIP( context: RegexLexerContext ): void
		{
			context.token = null;
			context.skip = true;
		}
		
		public static function START_EAT( context: RegexLexerContext ): void
		{
			++eating;
		}
		
		public static function STOP_EAT( context: RegexLexerContext ): void
		{
			--eating;
			
			context.token = null;
			context.skip = true;
		}
		
		public static function NL( context: RegexLexerContext ): void
		{
			context.token = null;
			context.skip = true;
			context.lineno += 1;
		}
	}
}

class LexerMatchRule
{
	public var type: int;
	public var index: int;
	public var name: String;
	public var subpatterns: int;
	public var pattern: String;
	public var callback: Function;
}